/*
								+---------------------------------+
								|                                 |
								|  ***   Lightmap computer   ***  |
								|                                 |
								|  Copyright   -tHE SWINe- 2003  |
								|                                 |
								|           Lightmap.cpp          |
								|                                 |
								+---------------------------------+
*/

#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "Matrix.h"
#include "font_eng.h"
#include "Load_3ds.h"
#include "Lightmap.h"
#include "radiosity.h"

#define _Len(A) ((float)sqrt(A.x * A.x + A.y * A.y + A.z * A.z))
// spot dlku vektoru

#define _Normalize(A) {float s = _Len(A); A.x /= s; A.y /= s; A.z /= s; }
// vytvo jednotkov vektor

#define _Dot(A,B) (B.x * A.x + A.y * B.y + A.z * B.z)
// skalrn souin

#define _Cross(A,B,C) C.x = B.y * A.z - A.y * B.z; \
                     C.y = B.z * A.x - A.z * B.x; \
                     C.z = B.x * A.y - A.x * B.y;
#define _CrossP(A,B,C) C.x = B.b * A.z - A.y * B.c; \
                     C.y = B.c * A.x - A.z * B.a; \
                     C.z = B.a * A.y - A.x * B.b;
// vektorov souin

const int n_lightmap_size = 512;
const float f_quality = .5f;

Object *p_IsolateVertices(Object *p_object)
{
	Object *p_new_object;
	int i, j;

	if(!(p_new_object = (Object*)malloc(sizeof(Object))))
		return 0;
	//
	p_new_object->n_face_num = p_object->n_face_num;
	p_new_object->n_vertex_num = p_object->n_face_num * 3;
	//
	if(!(p_new_object->face = (Face*)malloc(sizeof(Face) * p_object->n_face_num)))
		return 0;
	if(!(p_new_object->vertex = (Vertex*)malloc(sizeof(Vertex) * p_new_object->n_vertex_num)))
		return 0;
	// naalokuje objekt, vertexy a facy

	for(i = 0; i < p_object->n_vertex_num; i ++)
		p_object->vertex[i].n_id = i;
	// ulom si do kadho vertexu jeho id. podle
	// toho potom budu zase vertexy dvat dohromady ...

	for(i = 0; i < p_object->n_face_num; i ++) {
		p_new_object->face[i] = p_object->face[i];
		//
		for(j = 0; j < 3; j ++) {
			p_new_object->vertex[i * 3 + j] = *p_object->face[i].vertex[j];
			p_new_object->face[i].vertex[j] = &p_new_object->vertex[i * 3 + j];
		}
		// zkopruje face a pehod pointery na vertexy, jejich obsah taky zkopruje
	}

	return p_new_object;
}

// vrt index polka ...
inline int n_Index(int cx, int cy, int x, int y, int w, int h, int pw)
{
	if(cx < x)
		cx = x;
	else if(cx >= x + w)
		cx = x + w - 1;
	if(cy < y)
		cy = y;
	else if(cy >= y + h)
		cy = y + h - 1;

	return cx + cy * pw;
}

int Blur_Window(int x, int y, int w, int h, int pw, int n_size, float *p_buffer, int n_offset)
{
	float f_r, f_g, f_b, f_size, f_weight;
	float *p_temp_buffer;
	int cx, cy;
	int u, v;

	if(!(p_temp_buffer = (float*)malloc(w * h * sizeof(float) * 3)))
		return false;

	for(v = y; v < y + h; v ++) {
		for(u = x; u < x + w; u ++) {
			f_r = 0;
			f_g = 0;
			f_b = 0;
			f_size = 0;
			//
			for(cx = u - n_size; cx <= u + n_size; cx ++) {
				for(cy = v - n_size; cy <= v + n_size; cy ++) {
					f_weight = 1 / (1 + (float)sqrt((cx - u) * (cx - u) + (cy - v) * (cy - v)));
					//
					f_r += p_buffer[n_Index(cx, cy, x, y, w, h, pw) * 6 + 0 + n_offset] * f_weight;
					f_g += p_buffer[n_Index(cx, cy, x, y, w, h, pw) * 6 + 1 + n_offset] * f_weight;
					f_b += p_buffer[n_Index(cx, cy, x, y, w, h, pw) * 6 + 2 + n_offset] * f_weight;

					f_size += f_weight;
				}
			}
			//
			f_size = 1 / f_size;

			f_r *= f_size;
			f_g *= f_size;
			f_b *= f_size;

			p_temp_buffer[(u - x + (v - y) * w) * 3] = f_r;
			p_temp_buffer[(u - x + (v - y) * w) * 3 + 1] = f_g;
			p_temp_buffer[(u - x + (v - y) * w) * 3 + 2] = f_b;
		}
	}
	// rozmae obrzek do doasnhu bufferu ...

	for(v = y; v < y + h; v ++) {
		for(u = x; u < x + w; u ++) {
			p_buffer[(u + v * pw) * 6 + 0 + n_offset] = p_temp_buffer[(u - x + (v - y) * w) * 3];
			p_buffer[(u + v * pw) * 6 + 1 + n_offset] = p_temp_buffer[(u - x + (v - y) * w) * 3 + 1];
			p_buffer[(u + v * pw) * 6 + 2 + n_offset] = p_temp_buffer[(u - x + (v - y) * w) * 3 + 2];
		}
	}
	// zkopruje data z doasnhu bufferu ...

	free((void*)p_temp_buffer);

	return true;
}

int Create_Lightmaps(char *p_s_filename, Mesh *p_mesh)
{
	TLightmapTex **p_lightmap_page;
	int n_lightmap_page_num;
	Mapping_Info *p_m_info;
	Matrix m_coord_trans;
	Group_Data *p_g_data;
	int n_cur_w, n_cur_h;
	int n_cur_x, n_cur_y;
	float f_tr_u, f_tr_v;
	int i, j, k, l, n;
	Object *p_object;
	World *p_svet;
	Vector v_temp, r;

	if(!(p_svet = Load_3DS(p_s_filename)))
		return false;
	// nahraje 3ds

	k = 0;

	for(i = 0; i < p_svet->n_object_num; i ++) {
		for(j = 0; j < p_svet->object[i].n_face_num; j ++) {
			if(!stricmp(p_svet->material[p_svet->object[i].face[j].n_material].name, "wht")) {
				v_temp.x = (p_svet->object[i].face[j].vertex[0]->x +
							p_svet->object[i].face[j].vertex[1]->x +
							p_svet->object[i].face[j].vertex[2]->x) / 3;
				v_temp.y = (p_svet->object[i].face[j].vertex[0]->y +
							p_svet->object[i].face[j].vertex[1]->y +
							p_svet->object[i].face[j].vertex[2]->y) / 3;
				v_temp.z = (p_svet->object[i].face[j].vertex[0]->z +
							p_svet->object[i].face[j].vertex[1]->z +
							p_svet->object[i].face[j].vertex[2]->z) / 3;
				//
				f_tr_u = 1e10;
				if(!p_svet->n_light_num) {
					p_svet->object[i].face[j].f_radiative_r = __energy_base__;
					p_svet->object[i].face[j].f_radiative_g = __energy_base__;
					p_svet->object[i].face[j].f_radiative_b = __energy_base__;
				}
				for(n = 0; n < p_svet->n_light_num; n ++) {
					r.x = v_temp.x - p_svet->light[n].v_position.x;
					r.y = v_temp.y - p_svet->light[n].v_position.y;
					r.z = v_temp.z - p_svet->light[n].v_position.z;

					if(_Len(r) < f_tr_u) {
						f_tr_u = _Len(r);
						//
						p_svet->object[i].face[j].f_radiative_r = __energy_base__ *
							p_svet->light[n].r * p_svet->light[n].f_multiplier;
						p_svet->object[i].face[j].f_radiative_g = __energy_base__ *
							p_svet->light[n].g * p_svet->light[n].f_multiplier;
						p_svet->object[i].face[j].f_radiative_b = __energy_base__ *
							p_svet->light[n].b * p_svet->light[n].f_multiplier;

						break;
					}
				}
				// ur barvu radiativn energie fac podle nejbliho svtla ...
			} else {
				p_svet->object[i].face[j].f_radiative_r = 0;
				p_svet->object[i].face[j].f_radiative_g = 0;
				p_svet->object[i].face[j].f_radiative_b = 0;
			}
			p_svet->object[i].face[j].n_id = k ++;
		}
	}
	// facm svtel piad radiativn energii ...

	for(i = 0; i < p_svet->n_object_num; i ++) {
		if(!(p_object = p_IsolateVertices(&p_svet->object[i])))
			return false;
		// vytvo objekt s izolovanmi vertexy (kad face m svoje vlastn vertexy)

		free((void*)p_svet->object[i].face);
		free((void*)p_svet->object[i].vertex);
		// uvoln stvajc vertexy a facy

		p_svet->object[i].face = p_object->face;
		p_svet->object[i].n_vertex_num = p_object->n_vertex_num;
		p_svet->object[i].vertex = p_object->vertex;
		// nastav nov vertexy a jejich poet a facy (poet fac se nezmn)

		free((void*)p_object);
		// uvoln pomocn objekt
	}
	// izoluje vertexy ...

	if(!(p_g_data = Group_Faces(p_svet, (float)n_lightmap_size * f_quality)))
		return false;
	// vytvo groupy fac, kad group bude mt jednu lightmapu

	if(!(p_m_info = Generate_Map_Coords(p_g_data)))
		return false;
	// spot mapovac souadnice (u, v pro lightmapy)

	n_lightmap_page_num = 1;
	if(!(p_lightmap_page = (TLightmapTex**)malloc(sizeof(TLightmapTex*))))
		return false;
	// naalokuje pam pro list se strnkami ...
	
	if(!(p_lightmap_page[0] = p_Lightmap(n_lightmap_size, n_lightmap_size)))
		return false;
	// naalokuje prvn strnku ...

	for(i = 0; i < p_g_data->n_group_num; i ++) {
		n_cur_w = (int)ceil(p_m_info[i].f_width * f_quality) + 1;
		n_cur_h = (int)ceil(p_m_info[i].f_height * f_quality) + 1;
		//
		n_cur_w = (n_cur_w > n_lightmap_size)? n_lightmap_size : n_cur_w;
		n_cur_h = (n_cur_h > n_lightmap_size)? n_lightmap_size : n_cur_h;
		// ur velikosti mapy pro group i

		for(j = 0; j < n_lightmap_page_num; j ++) {
			if(b_InsertTile(n_cur_w, n_cur_h, p_lightmap_page[j],
			   &n_cur_x, &n_cur_y, &m_coord_trans)) {
				p_g_data->p_group[i].t_rect.x = n_cur_x;
				p_g_data->p_group[i].t_rect.y = n_cur_y;
				p_g_data->p_group[i].t_rect.w = n_cur_w;
				p_g_data->p_group[i].t_rect.h = n_cur_h;
				p_g_data->p_group[i].t_rect.n_page = j;

				for(k = 0; k < p_g_data->p_group[i].n_face_used; k ++) {
					for(l = 0; l < 3; l ++) {
						f_tr_u = m_coord_trans[0][0] *
								 p_g_data->p_group[i].p_face[k]->vertex[l]->lu +
								 m_coord_trans[0][1] *
								 p_g_data->p_group[i].p_face[k]->vertex[l]->lv +
								 m_coord_trans[3][0];
						f_tr_v = m_coord_trans[1][0] *
								 p_g_data->p_group[i].p_face[k]->vertex[l]->lu +
								 m_coord_trans[1][1] *
								 p_g_data->p_group[i].p_face[k]->vertex[l]->lv +
								 m_coord_trans[3][1];
						//
						p_g_data->p_group[i].p_face[k]->vertex[l]->lu = f_tr_u;
						p_g_data->p_group[i].p_face[k]->vertex[l]->lv = f_tr_v;
						//
						p_g_data->p_group[i].p_face[k]->n_lightmap = j;
					}
				}
				// natransformuje souadnice z naeho pomyslnho obdlnku do texturov strnky

				Raytrace_Tile(&p_g_data->p_group[i], &p_m_info[i], p_svet,
					n_cur_x, n_cur_y, n_cur_h, n_cur_w, p_lightmap_page[j]->n_width,
					p_lightmap_page[j]->p_bytes);
				// vypot svtlost jednoho dlku pmo v lightmap ...

				Blur_Window(n_cur_x, n_cur_y, n_cur_w, n_cur_h, p_lightmap_page[j]->n_width,
					1, p_lightmap_page[j]->p_bytes, 0);
				Blur_Window(n_cur_x, n_cur_y, n_cur_w, n_cur_h, p_lightmap_page[j]->n_width,
					1, p_lightmap_page[j]->p_bytes, 3);
				// rozmae radiativn a iluminativn energie ...

				for(n = 0; n < 6; n ++) {
					for(k = n_cur_x; k < n_cur_x + n_cur_w; k ++) {
						p_lightmap_page[j]->p_bytes[((k +
							p_lightmap_page[j]->n_width * (n_cur_y - 1)) * 6) + n] =
							p_lightmap_page[j]->p_bytes[((k +
							p_lightmap_page[j]->n_width * n_cur_y) * 6) + n];
						p_lightmap_page[j]->p_bytes[((k +
							p_lightmap_page[j]->n_width * (n_cur_y + n_cur_h)) * 6) + n] =
							p_lightmap_page[j]->p_bytes[((k +
							p_lightmap_page[j]->n_width * (n_cur_y + n_cur_h - 1)) * 6) + n];
					}
					// zkopruje rmeek horizontln ...

					for(l = n_cur_y - 1; l < n_cur_y + n_cur_h + 1; l ++) {
						p_lightmap_page[j]->p_bytes[((n_cur_x - 1 +
							p_lightmap_page[j]->n_width * l) * 6) + n] =
							p_lightmap_page[j]->p_bytes[((n_cur_x +
							p_lightmap_page[j]->n_width * l) * 6) + n];
						p_lightmap_page[j]->p_bytes[((n_cur_x + n_cur_w +
							p_lightmap_page[j]->n_width * l) * 6) + n] =
							p_lightmap_page[j]->p_bytes[((n_cur_x - 1 + n_cur_w +
							p_lightmap_page[j]->n_width * l) * 6) + n];
					}
					// zkopruje rmeek vertikln
				}
			} else if(j == n_lightmap_page_num - 1) {
				if(!(p_lightmap_page = (TLightmapTex**)realloc(p_lightmap_page,
				   ++ n_lightmap_page_num * sizeof(TLightmapTex*))))
					return false;
				//
				if(!(p_lightmap_page[n_lightmap_page_num - 1] =
				   p_Lightmap(n_lightmap_size, n_lightmap_size)))
					return false;
				// naalokuje dal strnku, pokud se textura nevejde do dn z pedchozch 
			}
		}
		// pid lightmapu do texturov strnky
	}

	p_mesh->n_lightmap_num = n_lightmap_page_num;
	p_mesh->n_object_num = p_svet->n_object_num;
	p_mesh->n_material_num = p_svet->n_material_num;
	//
	p_mesh->p_material = p_svet->material;
	p_mesh->p_object = p_svet->object;
	//
	if(!(p_mesh->p_lightmap = (Lightmap*)malloc(p_mesh->n_lightmap_num * sizeof(Lightmap))))
		return false;

	for(i = 0; i < p_mesh->n_lightmap_num; i ++) {
		p_mesh->p_lightmap[i].n_width = p_lightmap_page[i]->n_width;
		p_mesh->p_lightmap[i].n_height = p_lightmap_page[i]->n_height;
		p_mesh->p_lightmap[i].p_buffer = p_lightmap_page[i]->p_bytes;
		//
		p_mesh->p_lightmap[i].p_s_filename = 0;
		//
		p_mesh->p_lightmap[i].A_1 = p_mesh->p_lightmap[i].n_width - 1;
		p_mesh->p_lightmap[i].A_2 = p_mesh->p_lightmap[i].n_width *
			(p_mesh->p_lightmap[i].n_height - 1);
		p_mesh->p_lightmap[i].A_3 = p_mesh->p_lightmap[i].n_width *
			p_mesh->p_lightmap[i].n_height - 1;
		p_mesh->p_lightmap[i].B_1 = 16 - (int)(log((double)p_mesh->p_lightmap[i].n_width) /
			log(2) + 1e-7);
		// zkopruje lightmapu

		p_lightmap_page[i]->p_bytes = 0;
		free(p_lightmap_page[i]->p_frontline);
		// uvoln zbytek struktury
	}
	free(p_lightmap_page);
	p_lightmap_page = 0;
	// zkopruje lightmapy

	p_mesh->p_m_info = p_m_info;
	p_mesh->p_g_data = p_g_data;
	// mapovac info a groupovac data ...

	p_svet->material = 0;
	p_svet->object = 0;
	free(p_svet->light);
	free(p_svet->camera);
	free(p_svet);
	// uvoln svt

	return true;
}

Vector v_Energy(float x, float y, float z)
{
	Vector v;

	v.x = x;
	v.y = y;
	v.z = z;

	return v;
}

float f_Energy(float x, float y, float z)
{
	return (x * x + y * y + z * z);
}

void Refine_Edges(Mesh *p_mesh)
{
	int n_cur_x, n_cur_y, n_cur_w, n_cur_h;
	int k, n, i, j;

	for(i = 0; i < p_mesh->p_g_data->n_group_num; i ++) {
		j = p_mesh->p_g_data->p_group[i].t_rect.n_page;
		n_cur_x = p_mesh->p_g_data->p_group[i].t_rect.x;
		n_cur_y = p_mesh->p_g_data->p_group[i].t_rect.y;
		n_cur_w = p_mesh->p_g_data->p_group[i].t_rect.w;
		n_cur_h = p_mesh->p_g_data->p_group[i].t_rect.h;

		for(n = 0; n < 3; n ++) {
			for(k = n_cur_x; k < n_cur_x + n_cur_w; k ++) {
				p_mesh->p_lightmap[j].p_buffer[((k +
					p_mesh->p_lightmap[j].n_width * (n_cur_y - 1)) * 6) + n] =
					p_mesh->p_lightmap[j].p_buffer[((k +
					p_mesh->p_lightmap[j].n_width * n_cur_y) * 6) + n];
				p_mesh->p_lightmap[j].p_buffer[((k +
					p_mesh->p_lightmap[j].n_width * (n_cur_y + n_cur_h)) * 6) + n] =
					p_mesh->p_lightmap[j].p_buffer[((k +
					p_mesh->p_lightmap[j].n_width * (n_cur_y + n_cur_h - 1)) * 6) + n];
			}
			// zkopruje rmeek horizontln ...

			for(k = n_cur_y - 1; k < n_cur_y + n_cur_h + 1; k ++) {
				p_mesh->p_lightmap[j].p_buffer[((n_cur_x - 1 +
					p_mesh->p_lightmap[j].n_width * k) * 6) + n] =
					p_mesh->p_lightmap[j].p_buffer[((n_cur_x +
					p_mesh->p_lightmap[j].n_width * k) * 6) + n];
				p_mesh->p_lightmap[j].p_buffer[((n_cur_x + n_cur_w +
					p_mesh->p_lightmap[j].n_width * k) * 6) + n] =
					p_mesh->p_lightmap[j].p_buffer[((n_cur_x - 1 + n_cur_w +
					p_mesh->p_lightmap[j].n_width * k) * 6) + n];
			}
			// zkopruje rmeek vertikln
		}
	}
}

void Blur_Illumination(Mesh *p_mesh, int n_strength)
{
	int n_cur_x, n_cur_y, n_cur_w, n_cur_h;
	int i, j;

	for(i = 0; i < p_mesh->p_g_data->n_group_num; i ++) {
		j = p_mesh->p_g_data->p_group[i].t_rect.n_page;
		n_cur_x = p_mesh->p_g_data->p_group[i].t_rect.x;
		n_cur_y = p_mesh->p_g_data->p_group[i].t_rect.y;
		n_cur_w = p_mesh->p_g_data->p_group[i].t_rect.w;
		n_cur_h = p_mesh->p_g_data->p_group[i].t_rect.h;

		Blur_Window(n_cur_x, n_cur_y, n_cur_w, n_cur_h, p_mesh->p_lightmap[j].n_width,
			n_strength, p_mesh->p_lightmap[j].p_buffer, 0);
	}
}

float f_Find_GreatestRadiativeEnergy(Mesh *p_mesh, Vector *v_texel_pos,
									 Vector *v_texel_normal, int *n, Vector *v_rad_color)
{
	float f_max_energy, f_energy;
	int n_max_j, n_max_page;
	int n_watchdog;
	Group *p_group;
	int x, y;
	int i, j;

	n_watchdog = 0;
	do {
		f_max_energy = -1;
		n_max_j = -1;
		n_max_page = -1;
		for(i = 0; i < p_mesh->n_lightmap_num; i ++) {
			for(j = 0; j < p_mesh->p_lightmap[i].n_width * p_mesh->p_lightmap[i].n_height; j ++) {
				if((f_energy = f_Energy(p_mesh->p_lightmap[i].p_buffer[(j * 6) + 3],
										p_mesh->p_lightmap[i].p_buffer[(j * 6) + 4],
										p_mesh->p_lightmap[i].p_buffer[(j * 6) + 5])) > f_max_energy) {
					f_max_energy = f_energy;
					n_max_j = j;
					n_max_page = i;
				}
			}
		}
		// najde maximln radiativn energii v njak lightmap ...

		if(n_max_j == -1)
			return -1;

		for(i = 0; i < p_mesh->p_g_data->n_group_num; i ++) {
			if(p_mesh->p_g_data->p_group[i].t_rect.n_page != n_max_page)
				continue;

			x = n_max_j % p_mesh->p_lightmap[n_max_page].n_width;
			y = n_max_j / p_mesh->p_lightmap[n_max_page].n_width;
			// souadnice pixelu ...

			p_group = &p_mesh->p_g_data->p_group[i];
			//
			x -= p_group->t_rect.x;
			y -= p_group->t_rect.y;
			//
			if(x >= 0 && x <= p_group->t_rect.w &&
			   y >= 0 && y <= p_group->t_rect.h) {
				// nali jsme group, do kterho pixel pat ...

				v_texel_pos->x = p_mesh->p_m_info[i].v_org.x +

					p_mesh->p_m_info[i].v_v.x * (float)(y + .5f) /
					(float)p_group->t_rect.h *
					p_mesh->p_m_info[i].f_height +

					p_mesh->p_m_info[i].v_u.x * (float)(x + .5f) /
					(float)p_group->t_rect.w *
					p_mesh->p_m_info[i].f_width;
				//
				v_texel_pos->y = p_mesh->p_m_info[i].v_org.y +

					p_mesh->p_m_info[i].v_v.y * (float)(y + .5f) /
					(float)p_group->t_rect.h *
					p_mesh->p_m_info[i].f_height +

					p_mesh->p_m_info[i].v_u.y * (float)(x + .5f) /
					(float)p_group->t_rect.w *
					p_mesh->p_m_info[i].f_width;
				//
				v_texel_pos->z = p_mesh->p_m_info[i].v_org.z +

					p_mesh->p_m_info[i].v_v.z * (float)(y + .5f) /
					(float)p_group->t_rect.h *
					p_mesh->p_m_info[i].f_height +

					p_mesh->p_m_info[i].v_u.z * (float)(x + .5f) /
					(float)p_group->t_rect.w *
					p_mesh->p_m_info[i].f_width;
				//
				switch(p_mesh->p_m_info[i].n_plane) {
				case Plane_xy:
					v_texel_pos->z = -(v_texel_pos->x * p_group->p_face[0]->normal.a +
									  v_texel_pos->y * p_group->p_face[0]->normal.b +
									  p_group->p_face[0]->normal.d) / p_group->p_face[0]->normal.c;
					break;
				case Plane_xz:
					v_texel_pos->y = -(v_texel_pos->x * p_group->p_face[0]->normal.a +
									  v_texel_pos->z * p_group->p_face[0]->normal.c +
									  p_group->p_face[0]->normal.d) / p_group->p_face[0]->normal.b;
					break;
				case Plane_yz:
					v_texel_pos->x = -(v_texel_pos->y * p_group->p_face[0]->normal.b +
									  v_texel_pos->z * p_group->p_face[0]->normal.c +
									  p_group->p_face[0]->normal.d) / p_group->p_face[0]->normal.a;
					break;
				}
				// spot pozici texelu ve svt ...

				*n = p_group->p_face[0]->n_id;
				// id facu

				v_texel_normal->x = p_group->p_face[0]->normal.a;
				v_texel_normal->y = p_group->p_face[0]->normal.b;
				v_texel_normal->z = p_group->p_face[0]->normal.c;
				// normla texelu

				*v_rad_color = v_Energy(
					p_mesh->p_lightmap[n_max_page].p_buffer[(n_max_j * 6) + 3],
					p_mesh->p_lightmap[n_max_page].p_buffer[(n_max_j * 6) + 4],
					p_mesh->p_lightmap[n_max_page].p_buffer[(n_max_j * 6) + 5]);
				// vrt radiativn energii ...

				p_mesh->p_lightmap[n_max_page].p_buffer[(n_max_j * 6) + 3] = 0;
				p_mesh->p_lightmap[n_max_page].p_buffer[(n_max_j * 6) + 4] = 0;
				p_mesh->p_lightmap[n_max_page].p_buffer[(n_max_j * 6) + 5] = 0;
				// odstran radiativn energii ...

				return f_max_energy;
			}
		}

		p_mesh->p_lightmap[n_max_page].p_buffer[(n_max_j * 6) + 3] = 0;
		p_mesh->p_lightmap[n_max_page].p_buffer[(n_max_j * 6) + 4] = 0;
		p_mesh->p_lightmap[n_max_page].p_buffer[(n_max_j * 6) + 5] = 0;
		// vznikla chyba rasterizru a radiace se vykreslila kousek mimo face,
		// take nikam nepat ... odstranme ji a zkusme to s dal ...
	} while(n_watchdog ++ < 1024);

	return -1;
}
