/*
								+---------------------------------+
								|                                 |
								|   ***   S_Buffer engine   ***   |
								|                                 |
								|  Copyright   -tHE SWINe- 2002  |
								|                                 |
								|            Main.cpp             |
								|                                 |
								+---------------------------------+
*/

#include <windows.h>
#include <malloc.h>
#include <stdio.h>
#include <math.h>
#include "Directx.h"
#include "Matrix.h"
#include "Font_eng.h"
#include "Load_3ds.h"
#include "lightmap.h"
#include "Raster.h"
#include "radiosity.h"
#include "Main.h"

#define Pi (3.1415926535897932384626433832795028841931f)

#define APPNAME "lightmap-based radiosity processor"
#define WNDNAME "lightmap-based radiosity processor"
static int Width = 640;
static int Height = 480;

unsigned __int32 *video;
unsigned __int32 *buffer, *buffer2;

static Mesh t_svet;
// svet

static Matrix cmatrix;
static Matrix _no_y_cmatrix;
static float y_angle;
static int mx, my, ax, ay;
// kamera + kramy pro mouselook

static int nFrame = 0;
// frame counter (pocitadlo snimku)

static int b_iterate = true;
static int b_screen = true;
static int n_cubemap = 2;
static int b_show_patch = false;
static int b_radiative = false;
static int b_hq = false;

const int f_min_radiation = 0;

// vytvori soubor engine.log
// a vlozi do nej uvodni radku
void Eng_Log_Init()
{
	FILE *fw;
	fw = fopen("Engine.log", "w");
	fprintf(fw, "__________BSP - Log__________\n");
	fclose(fw);
}

// pripoji k souboru engine.log text l_str a odradkuje
void Eng_Log(char *l_str)
{
	FILE *fw;
	fw = fopen("Engine.log", "a");
	fprintf(fw, "%s\n", l_str);
	fclose(fw);
}

// main
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdline, int nCmdShow)
{
	char szAppName[] = APPNAME;
	WNDCLASSEX wndclass;
	HWND hwnd;
	BMP *p_bmp;
	MSG msg;

	Eng_Log_Init();

	wndclass.cbSize			= sizeof(wndclass);
	wndclass.style			= CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc	= Wokno;
	wndclass.cbClsExtra		= 0;
	wndclass.cbWndExtra		= 0;
	wndclass.hInstance		= hInstance;
	wndclass.hIcon			= LoadIcon(hInstance, IDI_APPLICATION);
	wndclass.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground	= (HBRUSH)GetStockObject(BLACK_BRUSH);
	wndclass.lpszMenuName	= 0;
	wndclass.lpszClassName	= szAppName;
	wndclass.hIconSm		= LoadIcon(NULL, IDI_APPLICATION);
	RegisterClassEx(&wndclass);
	hwnd = CreateWindow(szAppName,
		WNDNAME, WS_DLGFRAME,
		CW_USEDEFAULT, CW_USEDEFAULT,
		Width, Height, NULL,
		NULL, hInstance, NULL);
	ShowWindow(hwnd, nCmdShow);
	UpdateWindow(hwnd);
	ShowCursor(FALSE);
	// vytvo okno

	Matrix_Init(&_no_y_cmatrix);
	Matrix_Rotate_Z(&_no_y_cmatrix, (float)-Pi);
	Matrix_Translate(&_no_y_cmatrix, 75, -30, -50);
	y_angle = 0;
	// kamera ok

	Calc_HemicubeFormFactors();
	// spot form faktory hemicube

	Save_TrueColor_BMP("hemicube.bmp", p_bmp = p_Hemicube_Test());
	free(p_bmp->bytes);
	free(p_bmp);
	p_bmp = 0;
	// ulo obrzek hemicube ...

	Eng_Log("Loading world ..");

	if(!Create_Lightmaps("sshock_room_1obj.3ds", &t_svet))
		return 0;
	// nahraje svet

	Eng_Log("    OK");

	Eng_Log("Loading textures ...");

	if(!Load_Textures(&t_svet, "Textures"))
		return 0;
	// nahraje textury

	Eng_Log("    OK");

	Eng_Log("Init engine ...");
	
	Init_Engine(Width, Height);
	Init_Font_Engine(Width, Height);
	// engine ok

	Eng_Log("    OK");

	InitDirectX(hwnd, Width, Height);
	// inicializovano okno

	Eng_Log("DDraw is running..");

	Eng_Log("Locking some memory ...");

	if(!(buffer = (unsigned __int32*) malloc(Width * Height * 4)))
		return 0;
	if(!(buffer2 = (unsigned __int32*) malloc(Width * Height * 4)))
		return 0;

	Eng_Log("    OK");

	Eng_Log("Rendering !");

	SetTimer(hwnd, 0, 20, NULL);

	SetCursorPos(Width >> 1, Height >> 1);
	mx = Width >> 1;
	my = Height >> 1;
	// pripravi MouseLook

	while (GetMessage (&msg, NULL, 0, 0)) {
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	}

	free((void*)buffer);
	// uvolnuje pamet
	Terminate_Engine();
	Terminate_Font_Engine();
	// uvolnuje pamet v jinych modulech

	return msg.wParam;
}

// ati technologies filtr quincunx
void Quincunx(unsigned __int32 *src, unsigned __int32 *dest)
{
	int w2, h2, w1;
	int x, y;

	w2 = Width - 1;
	h2 = Height - 1;
	w1 = w2 + 1;

	y = Width * Height;
	for(x = 0; x < y; x ++)
		src[x] = ((src[x] >> 2) & 0x3f3f3f);

	for(y = 0; y < h2; y ++, dest += Width, src += Width) {
		for(x = 0; x < w2; x ++)
			dest[x] = src[x] + src[x + 1] + src[x + w1] + src[x + w2];
		dest[x] = src[x];
	}
	memcpy(dest, src, Width << 2);

	// obrazovy filtr
}

void FullScreen_BlendAdd(unsigned __int32 *p_dest, unsigned __int32 *p_add)
{
	register int r, g;
	int i;

	for(i = 0; i < Width * Height; i ++) {
		if(!p_add[i])
			continue;

		//j = ((i % Width) >> 1) + 320 * ((i / Width) >> 1);

		r = (p_add[i] & 0xff0000) * 4 + (p_dest[i] & 0xff0000); r = (r > 0xff0000)? 0xff0000 : r;
		g = (p_add[i] & 0x00ff00) * 4 + (p_dest[i] & 0x00ff00); r |= (g > 0x00ff00)? 0x00ff00 : g;
		g = (p_add[i] & 0x0000ff) * 4 + (p_dest[i] & 0x0000ff); r |= (g > 0x0000ff)? 0x0000ff : g;

		p_dest[i] = r;
	}
}

#define __NO_MMX

void FullScreen_Blur(unsigned __int32 *p_src, unsigned __int32 *p_dest, int n_strength)
{
	unsigned __int32 pix;
#ifdef __NO_MMX
	int x, y, cx;
	register int r, g, b;
#else
	int x, y, r;
	register int cx;
#endif
	int n, w, m;

	for(y = 0; y < Height; y ++) {
		for(x = 0; x < Width; x ++) {
			n = 0;

#ifdef __NO_MMX
			r = g = b = 0;
#else
			__asm {
				pxor	mm3, mm3        ; mm3 = 0 0 0 0 0 0 0 0
				pxor    mm5, mm5        ; mm5 = 0 0 0 0 0 0 0 0
			}
#endif
			m = x + n_strength;
			for(cx = x - n_strength + 1; cx < m; cx ++) {
				n += (w = n_strength - abs((cx - x)));

				pix = p_src[((cx >= 0)? ((cx < Width)? cx : Width - 1) : 0) + y * Width];

				if(!pix)
					continue;

#ifdef __NO_MMX
				r += ((pix & 0xff0000) >> 8) * w;
				g += (pix & 0x00ff00) * w;
				b += (pix & 0x0000ff) * w;
#else
				__asm {
					push eax

					mov eax, dword ptr pix
					movd        mm1, eax		; mm1 = 0 0 0 0 aA aR aG aB

					mov eax, dword ptr w
					push bx

					mov ah, al					; ax = 0 0 0 w
					mov bx, ax
					shl eax, 16
					mov ax, bx					; ax = w w w w

					pop bx


					movd        mm2, eax		; mm1 = 0 0 0 0 aA aR aG aB

					pop eax

					punpcklbw   mm1, mm5        ; mm1 = 0 aA 0 aR 0 aG 0 aB
					punpcklbw   mm2, mm5        ; mm2 = 0 aA 0 aR 0 aG 0 aB
					pmullw      mm1, mm2

					paddw mm3, mm1
				}
#endif
			}
#ifdef __NO_MMX
			r /= n;
			r <<= 8;
			r &= 0xff0000;
			g /= n;
			g &= 0x00ff00;
			b /= n;
			b &= 0x0000ff;

			p_dest[x + Width * y] = r | g | b;
#else
			__asm {
				push eax

				psrlw		mm1, 7

				packuswb    mm1, mm1        ; mm1 = 0 0 0 0 rA rR rG rB
			    movd        eax, mm1        ; eax = rA rR rG rB

				mov r, eax

				pop eax

				emms		; ...
			}
			
			p_dest[x + Width * y] = r;
#endif
		}
	}

	for(y = 0; y < Height; y ++) {
		for(x = 0; x < Width; x ++) {
			n = 0;

#ifdef __NO_MMX
			r = g = b = 0;
#else
			__asm {
				pxor	mm3, mm3        ; mm3 = 0 0 0 0 0 0 0 0
				pxor    mm5, mm5        ; mm5 = 0 0 0 0 0 0 0 0
			}
#endif
			m = y + n_strength;
			for(cx = y - n_strength + 1; cx < m; cx ++) {
				n += (w = n_strength - abs((cx - y)));

				pix = p_dest[((cx >= 0)? ((cx < Height)? cx : Height - 1) : 0) * Width + x];
				if(!pix)
					continue;

#ifdef __NO_MMX
				r += ((pix & 0xff0000) >> 8) * w;
				g += (pix & 0x00ff00) * w;
				b += (pix & 0x0000ff) * w;
#else
				__asm {
					push eax

					mov eax, dword ptr pix
					movd        mm1, eax		; mm1 = 0 0 0 0 aA aR aG aB

					mov eax, dword ptr w
					push bx

					mov ah, al					; ax = 0 0 0 w
					mov bx, ax
					shl eax, 16
					mov ax, bx					; ax = w w w w

					pop bx


					movd        mm2, eax		; mm1 = 0 0 0 0 aA aR aG aB

					pop eax

					punpcklbw   mm1, mm5        ; mm1 = 0 aA 0 aR 0 aG 0 aB
					punpcklbw   mm2, mm5        ; mm2 = 0 aA 0 aR 0 aG 0 aB
					pmullw      mm1, mm2

					paddw mm3, mm1
				}
#endif
			}
#ifdef __NO_MMX
			r /= n;
			r <<= 8;
			r &= 0xff0000;
			g /= n;
			g &= 0x00ff00;
			b /= n;
			b &= 0x0000ff;

			p_src[x + Width * y] = r | g | b;
#else
			__asm {
				push eax

				psrlw		mm1, 7

				packuswb    mm1, mm1        ; mm1 = 0 0 0 0 rA rR rG rB
			    movd        eax, mm1        ; eax = rA rR rG rB

				mov r, eax

				pop eax

				emms		; ...
			}
			
			p_src[x + Width * y] = r;
#endif
		}
	}
}

// funkce, zpracovavajici zpravy
// pri sipkach otaci kamerou, ins/del
// hybe kamerou dopredu dozadu
// a kresli obraz
// pri esc. skonci
LRESULT CALLBACK Wokno (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	Vector v_tx_pos, v_tx_norm, v_tx_color;
	int n_tx_face = 0;
	static int j = 0;
	float f_tx_size;
	Vertex v_a, v_b;
	char str[256];
	Matrix m;
	POINT p;
	int i;

	switch (iMsg) {
	case WM_KEYDOWN :
		if(wParam == 27) {
			ShowCursor(TRUE);
			SendMessage(hwnd, WM_CLOSE, 0, 0);
			return 0;
		} else {
			switch (wParam){
			case VK_LEFT:
				Matrix_Translate(&_no_y_cmatrix, -5, 0, 0);
				break;
			case VK_RIGHT:
				Matrix_Translate(&_no_y_cmatrix, 5, 0, 0);
				break;
			case VK_UP:
				Matrix_Translate(&_no_y_cmatrix, 0, -5 *
					(float)sin(y_angle), 5 * (float)cos(y_angle));
				break;
			case VK_DOWN:
				Matrix_Translate(&_no_y_cmatrix, 0, 5 *
					(float)sin(y_angle), -5 * (float)cos(y_angle));
				// ty kosinusy a sinusy zpusobi ze kamera se muze hybat i nahoru / dolu
				// tudiz uz je to presne jako kdyz v Half-Life nebo kdekoliv date "noclip"
				// (pokud by se rotovala jedinna matice kamery okolo osy x, tak by se mohla
				// otacet v ose pohledu, takze za chvili je vzhuru nohama - pokud smejdite
				// mysi porad dokolecka - a to nechci(byt vzhuru nohama))
				break;

			case 'S':
				b_screen = !b_screen;
				break;

			case 'P':
				b_show_patch = !b_show_patch;
				break;

			case 'H':
				n_cubemap = 2;
				break;
			case 'G':
				n_cubemap = 1;
				break;
			case 'F':
				n_cubemap = 0;
				break;

			case 'Q':
				b_hq = !b_hq;
				break;

			case 'I':
				b_iterate = !b_iterate;
				break;

			case 'R':
				b_radiative = !b_radiative;
				break;

			case 'B':
				Blur_Illumination(&t_svet, 2);
				break;
			}
		} // posouvani kamery sipkama
		return 0;

	case WM_TIMER :
		GetCursorPos(&p);
		SetCursorPos(Width >> 1, Height >> 1);
		ax = p.x - mx;
		ay = p.y - my;
		Matrix_Rotate_Y(&_no_y_cmatrix, (float)ax * (float)3/900);
		y_angle -= (float)ay * (float)3/900;
		memcpy(&cmatrix, &_no_y_cmatrix, sizeof(Matrix));
		Matrix_Rotate_X(&cmatrix, y_angle);
		// osetri mouselook

		memset(buffer, 0, sizeof(unsigned __int32) * Width * Height);
		// vycisti obrazovku

		if(b_iterate) {
			f_tx_size = f_Find_GreatestRadiativeEnergy(&t_svet, &v_tx_pos,
				&v_tx_norm, &n_tx_face, &v_tx_color);

			if(f_tx_size > f_min_radiation) {
				if(b_show_patch) {
					v_a.x = v_tx_pos.x;
					v_a.y = v_tx_pos.y;
					v_a.z = v_tx_pos.z;

					v_b.x = v_tx_norm.x * f_tx_size * -1e-10f + v_tx_pos.x;
					v_b.y = v_tx_norm.y * f_tx_size * -1e-10f + v_tx_pos.y;
					v_b.z = v_tx_norm.z * f_tx_size * -1e-10f + v_tx_pos.z;
				}
				// pozice normly ve worldspace, pokud ji chceme zobrazit

				v_tx_norm.x *= -1;
				v_tx_norm.y *= -1;
				v_tx_norm.z *= -1;

				v_tx_pos.x += v_tx_norm.x * 2.5f;
				v_tx_pos.y += v_tx_norm.y * 2.5f;
				v_tx_pos.z += v_tx_norm.z * 2.5f;
				// posune pohled kousek od patche, jinak vznikaj docela chyby

				Set_RenderTarget(RenderTarget_Radiate);
				Render_Hemicube(v_tx_pos, v_tx_norm, &t_svet, v_tx_color);
				// vyrenderuje hemicube a zrove tm osvtl nahlen patche ...

				Refine_Edges(&t_svet);
				// uprav rmeky, take nebudou vznikat tmav hrany ...
			}
		}

		if(b_screen) {
			SubWindow(Width, Height);
			if(b_radiative) {
				Init_Screen();
				for(i = 0; i < t_svet.n_object_num; i ++) {
					DrawObject(&t_svet.p_object[i].m_object, &cmatrix,
						&t_svet.p_object[i], t_svet.p_material, t_svet.p_lightmap);
				}
				Set_RenderTarget(RenderTarget_Radiation);
				Release_Screen(buffer);
			} else {
				if(b_hq) {
					//SubWindow(320, 240);
					Init_Screen();
					for(i = 0; i < t_svet.n_object_num; i ++) {
						DrawObject(&t_svet.p_object[i].m_object, &cmatrix,
							&t_svet.p_object[i], t_svet.p_material, t_svet.p_lightmap);
					}
					Set_RenderTarget(RenderTarget_Illumination_High);
					Release_Screen_LinearBuff(buffer2);

					/*i = Width;
					k = Height;
					Width = 320;
					Height = 240;*/
					FullScreen_Blur(buffer2, buffer, 32);
					/*Width = i;
					Height = k;*/
					SubWindow(Width, Height);
				}
				
				Init_Screen();
				for(i = 0; i < t_svet.n_object_num; i ++) {
					DrawObject(&t_svet.p_object[i].m_object, &cmatrix,
						&t_svet.p_object[i], t_svet.p_material, t_svet.p_lightmap);
				}
				Set_RenderTarget(RenderTarget_Illumination);
				Release_Screen(buffer);

				if(b_hq)
					FullScreen_BlendAdd(buffer, buffer2);
			}
			// kresli svet
		}

		if(n_cubemap == 1)
			Copy_HemicubeToScreen(buffer + 10 + 10 * Width, Width);
		else if(n_cubemap == 2)
			Copy_HemicubeToScreen2(buffer + 10 + 10 * Width, Width);
		// krel hemicube do odrazu

		if(b_show_patch) {
			Matrix_Inverse(&cmatrix, &m);
			Matrix_Multiply(&m, &t_svet.p_object[0].m_object, &m);
			for(i = 0; i < 3; i ++) {
				Line(m, t_svet.p_object[0].face[n_tx_face].vertex[i],
					t_svet.p_object[0].face[n_tx_face].vertex[(i + 1) % 3], buffer, 0x4082d0);
			}
			Line(m, &v_a, &v_b, buffer, 0xd08240);
		}
		// uke patch, ze kterho se zrovna stl

		if(InitVideoBuffer(&video)) {
			///
			memcpy(video, buffer, 4 * Width * Height);
			//Quincunx(buffer, video);
			/// zkopruje obraz

			sprintf(str, "[%7.2f, %7.2f, %7.2f], %f",
				v_tx_pos.x, v_tx_pos.y, v_tx_pos.z, f_tx_size);
			String_Out_AddTr(str, 10, Height - 35, video);

			if(j % 10 == 1)
				sprintf(str, "%d.st Frame", j ++);
			else if(j % 10 == 2)
				sprintf(str, "%d.nd Frame", j ++);
			else if(j % 10 == 3)
				sprintf(str, "%d.rd Frame", j ++);
			else
				sprintf(str, "%d.th Frame", j ++);
			String_Out_AddTr(str, 10, Height - 20, video);
			// vypise cislo snimku
		
			FlipVideoBuffer();
		}
		return 0;

	case WM_DESTROY :
		TerminateDirectX();
		PostQuitMessage (0);
		return 0;
	}

	return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}





