/*
								+---------------------------------+
								|                                 |
								| ***   Radiosity computer   ***  |
								|                                 |
								|  Copyright   -tHE SWINe- 2004  |
								|                                 |
								|          Radiosity.cpp          |
								|                                 |
								+---------------------------------+
*/

#include <windows.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "Matrix.h"
#include "font_eng.h"
#include "Load_3ds.h"
#include "Lightmap.h"
#include "radiosity.h"
#include "raster.h"

#define _Len(A) ((float)sqrt(A.x * A.x + A.y * A.y + A.z * A.z))
// spot dlku vektoru

#define _Normalize(A) {float s = _Len(A); A.x /= s; A.y /= s; A.z /= s; }
// vytvo jednotkov vektor

#define _Dot(A,B) (B.x * A.x + A.y * B.y + A.z * B.z)
// skalrn souin

#define _Cross(A,B,C) C.x = B.y * A.z - A.y * B.z; \
                     C.y = B.z * A.x - A.z * B.x; \
                     C.z = B.x * A.y - A.x * B.y;
// vektorov souin

#define Pi (3.1415926535897932384626433832795028841931f)

static const int n_hemicube_size = 64;

static unsigned __int32 p_s_hemicube_buffer[n_hemicube_size * n_hemicube_size];
static unsigned __int32 p_s_hemicube_buffer_side[4][n_hemicube_size * n_hemicube_size / 2];
static float p_hemicube_formfactor[n_hemicube_size * n_hemicube_size];
static float p_hemicube_formfactor_side[n_hemicube_size * n_hemicube_size / 2];

static float f_hemicube_max = 1;

void Calc_HemicubeFormFactors()
{
	float f_half_pixel_width;
	float f_pixel_area;
	float dx, dy;
	int x, y;
	float f;

	f_half_pixel_width = (1.0f / n_hemicube_size);
	//
	f_pixel_area = (2.0f / n_hemicube_size);
	f_pixel_area *= f_pixel_area;
	// 2 je velikost cubemapy

	for(x = 0; x < n_hemicube_size; x ++) {
		for (y = 0; y < n_hemicube_size; y ++) {
			dx = ((x - n_hemicube_size / 2) / (n_hemicube_size / 2.0f)) + f_half_pixel_width;
			dy = ((y - n_hemicube_size / 2) / (n_hemicube_size / 2.0f)) + f_half_pixel_width;

			f = (dx * dx + dy * dy + 1);
			f *= f * Pi;

			p_hemicube_formfactor[x + y * n_hemicube_size] = f_pixel_area / f;
		}
	}
	// vrchni ...

	for(x = 0; x < n_hemicube_size; x ++) {
		for(y = 0; y < n_hemicube_size / 2; y ++) {
			dx = (x - n_hemicube_size / 2) / (n_hemicube_size / 2.0f) + f_half_pixel_width;
			dy = (n_hemicube_size / 2 - 1 - y) / (n_hemicube_size / 2.0f) + f_half_pixel_width;

			f = (dx * dx + dy * dy + 1);
			f *= f * Pi;

			p_hemicube_formfactor_side[x + y * n_hemicube_size] =
				(f_pixel_area * (dy + f_half_pixel_width)) / f;
		}
	}
	// strana

	for(x = 0, f_hemicube_max = 0; x < n_hemicube_size; x ++) {
		for(y = 0; y < n_hemicube_size / 2; y ++) {
			if(f_hemicube_max < p_hemicube_formfactor_side[x + y * n_hemicube_size])
				f_hemicube_max = p_hemicube_formfactor_side[x + y * n_hemicube_size];
		}
	}
	//
	for(x = 0; x < n_hemicube_size; x ++) {
		for(y = 0; y < n_hemicube_size; y ++) {
			if(f_hemicube_max < p_hemicube_formfactor[x + y * n_hemicube_size])
				f_hemicube_max = p_hemicube_formfactor[x + y * n_hemicube_size];
		}
	}
	// hled nejvt poloku ...

#define __SUM_TEST
#ifdef __SUM_TEST
		float f_sum;

		for(y = 0, f_sum = 0; y < n_hemicube_size; y ++) {
			for(x = 0; x < n_hemicube_size / 2; x ++)
			    f_sum += p_hemicube_formfactor_side[y + x * n_hemicube_size];
		}
		// strana

		f_sum *= 4;
		// strany jsou 4

		for(x = 0; x < n_hemicube_size; x ++) {
			for(y = 0; y < n_hemicube_size; y ++)
				f_sum += p_hemicube_formfactor[x + y * n_hemicube_size];
		}
		// vrch

		// tady musi byt f_sum == 1, m to vyjde 1.00435f
#endif
}

BMP *p_Hemicube_Test()
{
	BMP *p_temp;
	int x, y, n;

	if(!(p_temp = (BMP*)malloc(sizeof(BMP))))
		return 0;

	p_temp->n_Width = n_hemicube_size;
	p_temp->n_Height = n_hemicube_size + n_hemicube_size / 2;

	if(!(p_temp->bytes = (unsigned __int32*)malloc(p_temp->n_Width *
	   p_temp->n_Height * sizeof(unsigned __int32))))
		return 0;

	for(x = 0; x < n_hemicube_size; x ++) {
		for(y = 0; y < n_hemicube_size; y ++) {
			n = (int)(p_hemicube_formfactor[x + y * n_hemicube_size] / f_hemicube_max * 0xff);
			p_temp->bytes[x + y * n_hemicube_size] = RGB(n, n, n);
		}
	}
	for(x = 0; x < n_hemicube_size; x ++) {
		for(y = 0; y < n_hemicube_size / 2; y ++) {
			n = (int)(p_hemicube_formfactor_side[x + y * n_hemicube_size] / f_hemicube_max * 0xff);
			p_temp->bytes[x + (y + n_hemicube_size) * n_hemicube_size] = RGB(n, n, n);
		}
	}
	// vytvo grayscale bitmapu ...

	return p_temp;
}

void Render_Hemicube(Vector v_pos, Vector v_normal, Mesh *p_svet, Vector v_radiative)
{
	Vector v_up, v_right;
	Matrix m_camera;
	int i, j;

	Matrix_Init(&m_camera);

	if(fabs(v_normal.x) < .01 && fabs(v_normal.y) - 1 < .01 && fabs(v_normal.z) < .01) {
		v_up.x = 0; 
		v_up.y = 0; 
		v_up.z = 1;
	} else {
		v_up.x = 0; 
		v_up.y = 1; 
		v_up.z = 0;
	}
	// najde smyslupln vektor nahoru ...

	_Normalize(v_normal);
	_Cross(v_up, v_normal, v_right);
	_Cross(v_right, v_normal, v_up);
	_Normalize(v_right);
	_Normalize(v_up);

	Matrix_Init(&m_camera);
	//
	m_camera[0][0] = v_right.x;
	m_camera[0][1] = v_right.y;
	m_camera[0][2] = v_right.z;
	m_camera[1][0] = v_up.x;
	m_camera[1][1] = v_up.y;
	m_camera[1][2] = v_up.z;
	m_camera[2][0] = v_normal.x;
	m_camera[2][1] = v_normal.y;
	m_camera[2][2] = v_normal.z;
	m_camera[3][0] = v_pos.x;
	m_camera[3][1] = v_pos.y;
	m_camera[3][2] = v_pos.z;
	// kamera, kter se dv dopedu

	Set_FormFactors(p_hemicube_formfactor, v_radiative);

	SubWindow(n_hemicube_size, n_hemicube_size);
	Clipper(n_hemicube_size, n_hemicube_size);
	Init_Screen();
	for(i = 0; i < p_svet->n_object_num; i ++) {
		DrawObject(&p_svet->p_object[i].m_object, &m_camera,
			&p_svet->p_object[i], p_svet->p_material, p_svet->p_lightmap);
	}
	Release_Screen_LinearBuff(p_s_hemicube_buffer);
	// vykresl pohled patche do pedn roviny ...

	Matrix_Rotate_X(&m_camera, -Pi / 2);
	// oto kameru tak, aby byla rovnobn s rovinou ...

	Set_FormFactors(p_hemicube_formfactor_side, v_radiative);

	SubWindow(n_hemicube_size, n_hemicube_size);
	Clipper(n_hemicube_size, n_hemicube_size / 2);
	for(i = 0; i < 4; i ++) {
		Init_Screen();
		for(j = 0; j < p_svet->n_object_num; j ++) {
			DrawObject(&p_svet->p_object[j].m_object, &m_camera,
				&p_svet->p_object[j], p_svet->p_material, p_svet->p_lightmap);
		}
		Release_Screen_LinearBuff(p_s_hemicube_buffer_side[i]);
		// vykresl pohled patche do bon roviny ...

		Matrix_Rotate_Y(&m_camera, -Pi / 2);
		// oto kameru k dal rovin ...
	}
}

void memcpy_r(void *p_a, const void *p_b, int n_size)
{
	register int i;
	unsigned __int32 *pa, *pb;

	pb = ((unsigned __int32*)p_b) + n_size / 4 - 1;
	pa = (unsigned __int32*)p_a;
	for(i = 0; i < n_size; i += 4)
		*pa ++ = *pb --;
}

void Modulate_HemicubeBuffers()
{
	int i, j, k;

	for(i = 0; i < n_hemicube_size * n_hemicube_size; i ++) {
		j = (int)(p_hemicube_formfactor[i] / f_hemicube_max * 0xff);
		p_s_hemicube_buffer[i] =
			((((p_s_hemicube_buffer[i] & 0xff00ff) * j) & 0xff00ff00) |
			(((p_s_hemicube_buffer[i] & 0x00ff00) * j) & 0x00ff0000)) >> 8;
	}
	//
	for(k = 0; k < 4; k ++) {
		for(i = 0; i < n_hemicube_size * n_hemicube_size / 2; i ++) {
			j = (int)(p_hemicube_formfactor_side[i] / f_hemicube_max * 0xff);
			p_s_hemicube_buffer_side[k][i] =
				((((p_s_hemicube_buffer_side[k][i] & 0xff00ff) * j) & 0xff00ff00) |
				(((p_s_hemicube_buffer_side[k][i] & 0x00ff00) * j) & 0x00ff0000)) >> 8;
		}
	}
	// ztmav hemicube podle form factor ...
}

void Copy_HemicubeToScreen(unsigned __int32 *p_buffer, int n_buffer_w)
{
	int i, j;

	//Modulate_HemicubeBuffers();

	for(i = 0; i < n_hemicube_size / 2; i ++) {
		memcpy_r(p_buffer + n_hemicube_size / 2 + i * n_buffer_w,
			p_s_hemicube_buffer_side[2] + (n_hemicube_size / 2 - 1 - i) * n_hemicube_size,
			n_hemicube_size * sizeof(unsigned __int32));
	}

	for(i = 0; i < n_hemicube_size / 2; i ++) {
		memcpy(p_buffer + n_hemicube_size / 2 + n_buffer_w *
			(i + (int)(n_hemicube_size * 1.5)),
			p_s_hemicube_buffer_side[0] + i * n_hemicube_size,
			n_hemicube_size * sizeof(unsigned __int32));
	}

	for(i = 0; i < n_hemicube_size; i ++) {
		memcpy(p_buffer + n_hemicube_size / 2 + n_buffer_w * (i + n_hemicube_size / 2),
			p_s_hemicube_buffer + i * n_hemicube_size,
			n_hemicube_size * sizeof(unsigned __int32));
	}

	for(i = 0; i < n_hemicube_size; i ++) {
		for(j = 0; j < n_hemicube_size / 2; j ++) {
			p_buffer[j + n_buffer_w * (i + n_hemicube_size / 2)] =
				p_s_hemicube_buffer_side[1][i + (n_hemicube_size / 2 - 1 - j) * n_hemicube_size];
		}
	}

	for(i = 0; i < n_hemicube_size; i ++) {
		for(j = 0; j < n_hemicube_size / 2; j ++) {
			p_buffer[j + (int)(n_hemicube_size * 1.5) +
				n_buffer_w * (i + n_hemicube_size / 2)] =
				p_s_hemicube_buffer_side[3][n_hemicube_size - 1 - i + j * n_hemicube_size];
		}
	}
}

void Copy_HemicubeToScreen2(unsigned __int32 *p_buffer, int n_buffer_w)
{
	Vector v;
	int w, h;
	int x, y;
	float f;

	//Modulate_HemicubeBuffers();

	w = (int)(2 * n_hemicube_size);
	h = w;

	for(y = 0; y < h; y ++) {
		for(x = 0; x < w; x ++) {
			v.x = -(w - x * 2) / (float)w;
			v.y = -(h - y * 2) / (float)h;
			if((f = v.x * v.x + v.y * v.y) > 1)
				continue;
			v.z = (float)sqrt(1 - f);
			// sestroj vektor, koukajc do cubemapy ...

			if(v.z > fabs(v.x) && v.z > fabs(v.y)) {
				p_buffer[x + y * n_buffer_w] =
					p_s_hemicube_buffer[(int)((v.x / v.z * .5f + .5f) * n_hemicube_size) +
									    (int)((v.y / v.z * .5f + .5f) * n_hemicube_size) * n_hemicube_size];
			} else if(fabs(v.x) > fabs(v.y)) {
				if(v.x < 0) {
					p_buffer[x + y * n_buffer_w] =
						p_s_hemicube_buffer_side[1]
						[(int)((-v.y / v.x * .5f + .5f) * (n_hemicube_size - 1)) +
						 (int)((v.z / v.x * .5f + .5f) * (n_hemicube_size - 1)) * n_hemicube_size];
					// 1
				} else {
					p_buffer[x + y * n_buffer_w] =
						p_s_hemicube_buffer_side[3]
						[(int)((-v.y / v.x * .5f + .5f) * (n_hemicube_size - 1)) +
						 (int)((-v.z / v.x * .5f + .5f) * (n_hemicube_size - 1)) * n_hemicube_size];
					// 3
				}
				// x = 1 / 3
			} else {
				if(v.y < 0) {
					p_buffer[x + y * n_buffer_w] =
						p_s_hemicube_buffer_side[2]
						[(int)((v.x / v.y * .5f + .5f) * (n_hemicube_size - 1)) +
						 (int)((v.z / v.y * .5f + .5f) * (n_hemicube_size - 1)) * n_hemicube_size];
					// 2
				} else {
					p_buffer[x + y * n_buffer_w] =
						p_s_hemicube_buffer_side[0]
						[(int)((v.x / v.y * .5f + .5f) * (n_hemicube_size - 1)) +
						 (int)((-v.z / v.y * .5f + .5f) * (n_hemicube_size - 1)) * n_hemicube_size];
					// 0
				}
				// y = 0 / 2
			}
		}
	}
}